import sys
import traceback
from pathlib import Path
from typing import Any, Optional
from urllib.parse import quote

import laspy
from pydantic import ValidationError
from pydantic.fields import ModelField
from PyQt5.QtCore import QEventLoop, QLocale, QObject, Qt, QThread, QUrl, pyqtSignal
from PyQt5.QtGui import (
    QCloseEvent,
    QDesktopServices,
    QDoubleValidator,
    QIcon,
    QIntValidator,
)
from PyQt5.QtWidgets import (
    QComboBox,
    QDialog,
    QFileDialog,
    QMainWindow,
    QMessageBox,
    QWidget,
)

from three_d_fin import __about__
from three_d_fin.gui.expert_dlg import Ui_Dialog
from three_d_fin.gui.main_window import Ui_MainWindow
from three_d_fin.processing.abstract_processing import FinProcessing
from three_d_fin.processing.configuration import FinConfiguration


class ApplicationWorker(QObject):
    """Simple worker to handle FinProcessing in a dedicated QThread.

    To avoid to block the main UI thread, we run the processing
    in a dedicated thread. This class encapsulate the FinProcessing
    object to make it suitable for a QThread processing
    """

    finished = pyqtSignal()
    error = pyqtSignal(str, str)
    memory_error = pyqtSignal()

    def __init__(self, processing_object: FinProcessing, parent=None):
        """Construct the Worker.

        Parameters
        ----------
        processing_object : FinProcessing
            An implementation of the abstract FinProcessing class.
            it is responsible for the computing logic.
            Its process() method is triggered by the "compute" button of the GUI.
        parent : Optional[QWidget]
            An optional parent. Should be None if run as standalone but could
            be a parent QWidget from the base application if run as a plugin

        """
        super().__init__(parent)
        self.processing_object = processing_object

    def run(self):
        """Run the FinProcessing object.

        This method is called by the QThread.
        If one Exception is raised during the process, it emits
        the string (msg) of the Exception
        """
        try:
            self.processing_object.process()
        except MemoryError:
            self.memory_error.emit()
        except Exception as e:
            self.error.emit(str(e), traceback.format_exc())
        self.finished.emit()


class ExpertDialog(QDialog):
    """Create the expert / about dialog.

    For now, it's an empty shell but it is left as is for further
    programmatically added parameters.
    """

    def __init__(self, parent=None):
        """Init the dialog."""
        super().__init__(parent)
        self.ui = Ui_Dialog()
        self.ui.setupUi(self)


class Application(QMainWindow):
    """The GUI Application."""

    processing_object: FinProcessing

    event_loop: Optional[QEventLoop] = None

    file_externally_defined: bool

    cloud_fields: Optional[list[str]]

    def __init__(
        self,
        processing_object: FinProcessing,
        file_externally_defined: bool = False,
        cloud_fields: Optional[list[str]] = None,
        parent: Optional[QWidget] = None,
    ):
        """Construct the 3DFin GUI Application.

        Parameters
        ----------
        processing_object : FinProcessing
            An implementation of the abstract FinProcessing class.
            it is responsible for the computing logic.
            Its process() method is triggered by the "compute" button of the GUI.
        file_externally_defined : bool
            Whether or not the file/filename was already defined by a third party.
            if True, input_las input and buttons will be disabled.
        cloud_fields : Optional[list[str]]
            List of candidates fields for the Z0 field. If present (not None),
            the z0_entry will be turned into a dropdown menu. If present but void,
            height normalization radio buttons will be disabled.
        parent : Optional[QWidget]
            An optional parent. Should be None if run as standalone but could
            be a parent QWidget from the base application if run as a plugin

        """
        super().__init__(parent)
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)
        self.setWindowTitle(f"3DFin v{__about__.__version__}")
        self.processing_object = processing_object
        self.file_externally_defined = file_externally_defined
        self.cloud_fields = cloud_fields

        # Force current index to be 0, since QT creator could change when .ui file is saved
        self.ui.tabWidget.setCurrentIndex(0)
        self.setWindowIcon(QIcon(":/assets/assets/icon_window.ico"))

        # Click on the "documentation" button
        self.ui.documentation_link_btn.clicked.connect(self._show_documentation)

        # Click on the "Tutorial" button
        self.ui.tutorial_link_btn.clicked.connect(self._show_tutorial)

        # Click on the "?" button on the expert table
        self.ui.expert_info_btn.clicked.connect(self._show_expert_dialog)

        # Click on "input file" button
        self.ui.input_file_btn.clicked.connect(self._ask_input_file)

        # Click on "output dir" button
        self.ui.output_dir_btn.clicked.connect(self._ask_output_dir)

        # Click on compute button
        self.ui.compute_btn.clicked.connect(self._compute_clicked)

        # Connect is_normalized check signal
        self.ui.is_normalized_chk.toggled.connect(self._normalize_toggled)

        # Handle the case of a predefined and limited choice of cloud_fields
        if self.cloud_fields is not None:
            layout = self.ui.z0_name_in.parent().layout()
            field_combo = QComboBox(self)
            field_combo.addItems(self.cloud_fields)
            layout.replaceWidget(self.ui.z0_name_in, field_combo)
            self.ui.z0_name_in.setParent(None)
            self.ui.z0_name_in = field_combo

        # Handle the case where the input file is provided externally
        if self.file_externally_defined:
            self.ui.input_file_lbl.setDisabled(True)
            self.ui.input_file_lbl.setText("File already set by the application")
            self.ui.input_file_btn.setDisabled(True)
            self.ui.input_file_in.setDisabled(True)

        self._populate_fields()

    def _populate_fields(self) -> None:
        """Populate fields with default values, labels, tooltips based on FinConfiguration.

        Fields defined in the configuration.py and QT fields defined in the ui files have the same name,
        we take advantage of that.

        We use this convention:
            [parameter_name]_lbl = main label, populated by FieldInfo.name
            [parameter_name]_in = main input, populated by default value i.e. FieldInfo.default
            [parameter_name]_ht = input hint, populated by FieldInfo.extra.hint
            tooltip_text is inferred from the FieldInfo.description
        This works for most numeric fields, but this mapping won't work for few exceptions:
            z0_name can be a QComboBox and a QTextEdit
            input_file / output_file have their own way to define default values
            Binary fields can be either QRadioButton or QCheckBox
        """
        config_dict = self.processing_object.config.dict()
        for config_section in config_dict:
            for key_param, value_param in config_dict[config_section].items():
                # Default z0_name should match one of the supplied list if present.
                tooltip_text = FinConfiguration.field_tooltip(config_section, key_param)
                hint_text = FinConfiguration.field_hint(config_section, key_param)
                field_type = FinConfiguration.field_type(config_section, key_param)
                if key_param == "z0_name" and self.cloud_fields is not None:
                    if value_param in self.cloud_fields:
                        id_default = self.cloud_fields.index(value_param)
                        self.ui.z0_name_in.setCurrentIndex(id_default)
                    self.ui.z0_name_in.setToolTip(tooltip_text)
                    self.ui.z0_name_lbl.setToolTip(tooltip_text)
                # Fix a minor presentation issue when no file is defined
                elif key_param == "input_file" and value_param is None:
                    self.ui.input_file_in.setText("")
                    self.ui.input_file_lbl.setToolTip(tooltip_text)
                    self.ui.input_file_in.setToolTip(tooltip_text)
                elif key_param == "output_dir":
                    self.ui.output_dir_in.setText(str(value_param))
                    self.ui.output_dir_in.setToolTip(tooltip_text)
                    self.ui.output_dir_lbl.setToolTip(tooltip_text)
                elif key_param == "is_normalized":
                    self.ui.is_normalized_chk.setChecked(not value_param)
                    self.ui.is_normalized_chk.setToolTip(tooltip_text)
                elif key_param == "is_noisy":
                    self.ui.is_noisy_chk.setChecked(value_param)
                    self.ui.is_noisy_chk.setToolTip(tooltip_text)
                elif key_param == "export_txt":
                    self.ui.export_txt_rb_1.setChecked(value_param)
                    self.ui.export_txt_rb_2.setChecked(not value_param)
                    self.ui.export_txt_lbl.setToolTip(tooltip_text)
                else:  # regular "numeric" QTextEdit live here.
                    input_field = getattr(self.ui, key_param + "_in")
                    input_field.setText(str(value_param))
                    # Set basic validators for both float and int
                    # it only checks if input is a valid float/int
                    # other constraints like range are left to pydantic validators
                    if issubclass(field_type, float):
                        # Force locale to C for decimal separator
                        validator = QDoubleValidator()
                        validator.setLocale(QLocale.c())
                        input_field.setValidator(validator)
                    elif issubclass(field_type, int):
                        validator = QIntValidator()
                        validator.setLocale(QLocale.c())
                        input_field.setValidator(validator)
                    input_field.setToolTip(tooltip_text)
                    getattr(self.ui, key_param + "_lbl").setToolTip(tooltip_text)
                    getattr(self.ui, key_param + "_ht").setText(hint_text)

    def _show_expert_dialog(self) -> None:
        """Show the expert help/about dialog."""
        dialog = ExpertDialog(self)
        dialog.show()

    def _show_tutorial(self):
        """Show the tutorial.

        Open the tutorial in the default browser
        """
        QDesktopServices.openUrl(QUrl("https://github.com/3DFin/3DFin_Tutorial/"))

    def _show_documentation(self) -> None:
        """Show the documentation.

        Open the default browser to show the documentation / or prompt user to choose a PDF viewer.
        """
        try:
            base_path = Path(sys._MEIPASS)
        except Exception:
            base_path = Path(__file__).absolute().parents[1] / "documentation"
        QDesktopServices.openUrl(QUrl.fromLocalFile(str(Path(base_path / "documentation.pdf").resolve())))

    def _show_normalization_warning(self) -> None:
        """Show Dialog with normalization warning.

        Poll the area_warning indicator and if true, show an alert Dialog.
        """
        if self.processing_object.area_warning:
            # We make a custom QMessageBox in order to display a "Rich Text"
            # that allows us to embed the link to the tutorial
            msg_box = QMessageBox(self)
            msg_box.setIcon(QMessageBox.Icon.Warning)
            msg_box.setWindowTitle("Potential normalization error")
            msg_box.setTextFormat(Qt.TextFormat.RichText)
            msg_box.setText(
                "3DFin has detected a potential error in the terrain modelling.<br>"
                + 'This usually happens when the "cloth resolution" parameter '
                + "didn't fit well the terrain.<br><br>"
                + "Errors in the terrain modelling lead to poor height normalization "
                + "and errors in the results, which is undesirable. "
                + "You can learn to how check those potential errors and improve "
                + "your results quickly using this dedicated "
                + '<a href="https://github.com/3DFin/3DFin_Tutorial/tree/main#case-i---adjusting-dtm-interpolation-settings">tutorial section</a>'
            )
            msg_box.exec_()

    def _ask_input_file(self) -> None:
        """Ask for a proper input las file.

        Current selected file is checked for validity (existence and type)
        in order to setup the initial dir and the initial file in
        the dialog. If a file is selected then it is checked to be a valid
        Las file before adding its path to the related input field.
        the output directory is changed accordingly (default to input las file
        parent directory)
        """
        initial_path = Path(self.ui.input_file_in.text())
        is_initial_file = initial_path.exists() and initial_path.is_file()
        initial_dir = initial_path.parent.resolve() if is_initial_file else Path.home()
        las_file, _ = QFileDialog.getOpenFileName(
            self,
            "3DFin input file",
            str(initial_dir),
            "las files (*.las *.Las *.Laz *.laz)",
        )

        if las_file == "" or None:
            return

        try:
            laspy.open(las_file, read_evlrs=False)
        except laspy.LaspyException:
            QMessageBox.critical(self, "3DFin Error", "Invalid input file")
            return

        self.ui.input_file_in.setText(str(Path(las_file).resolve()))
        self.ui.output_dir_in.setText(str(Path(las_file).parent.resolve()))

    def _ask_output_dir(self) -> None:
        """Ask for a proper output directory."""
        initial_path = Path(self.ui.output_dir_in.text())
        has_valid_initial_dir = initial_path.exists() and initial_path.is_dir()
        initial_dir = initial_path if has_valid_initial_dir else Path.home()
        self.ui.output_dir_in.setText(str(initial_dir.resolve()))

        output_dir = QFileDialog.getExistingDirectory(self, "3DFin output directory", str(initial_dir))

        # If the dialog was not closed/canceled
        if output_dir != "" and not None:
            self.ui.output_dir_in.setText(str(Path(output_dir).resolve()))

    def _get_parameters(self) -> dict[str, dict[str, Any]]:
        """Get parameters from widgets and return them organized in a dictionary.

        Like _populate_fields(), parameters are collected taking advantage of
        the match between Pydantic Fields name defined in the configuration
        module and the name of the input fields in the QT ui file.

        Returns
        -------
        options : dict[str, dict[str, any]]
            Dictionary of parameters. It is organized following the
            3DFinconfig.ini file: Each parameters are sorted in a sub-dict
            ("basic", "expert", "advanced", "misc").

        """
        config_dict: dict[str, dict[str, Any]] = dict()
        for category_name, category_field in FinConfiguration.__fields__.items():
            category_dict: dict[str, Any] = dict()
            for key_param in category_field.type_().__fields__:
                if key_param == "z0_name" and self.cloud_fields is not None:
                    category_dict[key_param] = self.ui.z0_name_in.currentText()
                # When file is externally defined, force it to None to avoid validation errors
                elif key_param == "input_file" and self.file_externally_defined:
                    category_dict[key_param] = None
                elif key_param == "is_normalized":
                    category_dict[key_param] = not self.ui.is_normalized_chk.isChecked()
                elif key_param == "is_noisy":
                    category_dict[key_param] = self.ui.is_noisy_chk.isChecked()
                elif key_param == "export_txt":
                    category_dict[key_param] = self.ui.export_txt_rb_1.isChecked()
                else:
                    category_dict[key_param] = getattr(self.ui, key_param + "_in").text()
            config_dict[category_name] = category_dict
        return config_dict

    def _compute_clicked(self) -> None:
        """Validate I/O entries and run the processing callback."""
        params = self._get_parameters()

        # Define a local function in order to popup errors.
        def _show_error(error_msg: str) -> str:
            return QMessageBox.critical(self, "3DFin Error", error_msg)

        # Pydantic checks, we check the validity of the data.
        try:
            fin_config = FinConfiguration.parse_obj(params)
        except ValidationError as validation_errors:
            final_msg: str = "Invalid Parameters:\n\n"
            for error in validation_errors.errors():
                error_loc: list[str] = error["loc"]
                # Get the human readable value for the field by introspection
                # (stored in "title "attribute).
                field: ModelField = FinConfiguration.__fields__[error_loc[0]].type_().__fields__[error_loc[1]]
                title = field.field_info.title
                # Error text formatting.
                final_msg = final_msg + f"{title} \n"
                final_msg = final_msg + f"""\t -> {error["msg"]} \n"""
            _show_error(final_msg)
            return

        self.processing_object.set_config(fin_config)

        # Here we will check in an abstract way if the output could collide
        # with results of previous computations... and ask if we want to overwrite them.
        if self.processing_object.check_already_computed_data():
            overwrite = QMessageBox.question(
                self,
                "3DFin",
                "The output target already contains results from a previous 3DFin computation, do you want to overwrite them?",
            )
            if overwrite == QMessageBox.No:
                return

        # Handle changes in the GUI when compute is launched/finished
        def _disable_btn() -> None:
            self.ui.compute_btn.setDisabled(True)
            self.ui.compute_btn.setText("Computing...")

        def _enable_btn() -> None:
            self.ui.compute_btn.setDisabled(False)
            self.ui.compute_btn.setText("Compute")

        def _memory_error_handling():
            _enable_btn()
            msg_box = QMessageBox(self)
            msg_box.setIcon(QMessageBox.Icon.Critical)
            msg_box.setWindowTitle("3DFin memory Error")
            msg_box.setTextFormat(Qt.TextFormat.RichText)
            msg_box.setText("3DFin encountered a memory error")
            msg_box.setInformativeText(
                "Try to free memory by closing other applications running on your computer.\n"
                + "You could also split your point cloud in smaller part or run 3DFin on a computer with a larger amount of RAM."
            )
            msg_box.exec_()

        def _error_handling(error_message: str, stack_trace: str) -> None:
            _enable_btn()
            msg_box = QMessageBox(self)
            msg_box.setIcon(QMessageBox.Icon.Critical)
            msg_box.setWindowTitle("3DFin Runtime Error")
            msg_box.setTextFormat(Qt.TextFormat.RichText)
            msg_box.setText(error_message)
            msg_box.setInformativeText(
                "For more details please read the stack trace below. "
                + "If you think this is a bug, please open an "
                + 'issue on <a href="https://github.com/3DFin/3DFin/issues/new?body='
                + quote(f"Hi!\n 3DFin v{__about__.__version__} failed with this error: \n {stack_trace}")
                + '">Github</a>'
            )
            msg_box.setDetailedText(stack_trace)
            msg_box.exec_()

        # Now we do the processing in itself
        self.thread = QThread()
        _disable_btn()
        # Pre processing hook is called from the main thread
        self.processing_object._pre_processing_hook()
        # Create a worker object
        self.worker = ApplicationWorker(self.processing_object)

        # Move the worker to the thread
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        # Post processing hook is called from the main thread
        self.worker.finished.connect(self.processing_object._post_processing_hook)
        self.worker.finished.connect(self._show_normalization_warning)
        self.worker.finished.connect(_enable_btn)
        self.worker.error.connect(_error_handling)
        self.worker.memory_error.connect(_memory_error_handling)
        self.thread.finished.connect(self.thread.deleteLater)

        self.thread.start()

    def _normalize_toggled(self) -> None:
        """Handle the 'is_normalized' checkbox toggle event."""
        self.ui.is_noisy_chk.setEnabled(self.ui.is_normalized_chk.isChecked())
        self.ui.z0_name_in.setEnabled(not self.ui.is_normalized_chk.isChecked())
        self.ui.z0_name_lbl.setEnabled(not self.ui.is_normalized_chk.isChecked())

    def set_event_loop(self, loop: QEventLoop) -> None:
        """Set an optional event loop.

        In some context (e.g. CloudCompare and QGIS plugin),
        we need to set a dedicated event loop to the mainwindow.

        Parameters
        ----------
        loop : QEventLoop
            The event loop to set

        """
        self.event_loop = loop

    def closeEvent(self, a0: QCloseEvent) -> None:
        """Close the application.

        The event loop is exited if it was previously set.

        Parameters
        ----------
        a0 : QCloseEvent
            The close event

        """
        super().closeEvent(a0)
        if self.event_loop is not None:
            self.event_loop.quit()



import sys
from pathlib import Path

import numpy as np
import pycc

from three_d_fin.cloudcompare.plugin_progress import CloudCompareProgress
from three_d_fin.processing.abstract_processing import FinProcessing
from three_d_fin.processing.configuration import FinConfiguration


class CloudComparePluginProcessing(FinProcessing):
    """Implement the FinProcessing interface for CloudCompare in a plugin context."""

    base_group: pycc.ccHObject

    group_name: str

    @staticmethod
    def write_sf(point_cloud: pycc.ccPointCloud, scalar_field: np.ndarray, name: str):
        """Write a scalar field on a pycc.PointCloud.

        Parameters
        ----------
        point_cloud : pycc.ccPointCloud
            Point cloud targetted by the 3DFin processing.
        scalar_field : np.ndarray
            Numpy vector discribing the scalar field.
        name: str
            Name of the scalar_field to write.

        """
        idx_sf = point_cloud.addScalarField(name)
        sf_array = point_cloud.getScalarField(idx_sf).asArray()
        sf_array[:] = scalar_field.astype(np.float32)[:]
        point_cloud.getScalarField(idx_sf).computeMinAndMax()

    def __init__(
        self,
        cc_instance: pycc.ccPythonInstance,
        point_cloud: pycc.ccPointCloud,
        config: FinConfiguration,
    ):
        """Construct the functor object.

        Parameters
        ----------
        cc_instance : pycc.ccPythonInstance
            Current cc application, wrapped by CloudCompare-PythonRuntime.
        point_cloud : pycc.ccPointCloud
            Point cloud targetted by the 3DFin processing.
        config: FinConfiguration
            The FinConfiguration object

        """
        self.base_cloud = point_cloud
        self.cc_instance = cc_instance
        self.progress = CloudCompareProgress(output=sys.stdout)
        super().__init__(config)

    def _construct_output_path(self):
        # We still use the stem attribute since in CC cloud could name could be based on filenames
        self.output_basepath = Path(self.config.misc.output_dir) / Path(self.base_cloud.getName()).stem
        self.group_name = f"{Path(self.base_cloud.getName()).stem}_3DFin"

    def _pre_processing_hook(self):
        # Be sure to load our custom colorscale
        color_scale_path = str((Path(__file__).parents[0] / "assets" / "3dfin_color_scale.xml").resolve())
        scale_manager = pycc.ccColorScalesManager.GetUniqueInstance()
        color_scale = pycc.ccColorScale.LoadFromXML(color_scale_path)
        scale_manager.addScale(color_scale)
        self.base_group = pycc.ccHObject(self.group_name)
        self.base_cloud.setEnabled(False)

    def _post_processing_hook(self):
        # Could be used to delay addToDB calls.
        self.cc_instance.addToDB(self.base_group)

    def _load_base_cloud(self):
        # This is already loaded at object instanciation.
        pass

    def _get_xyz_z0_from_base(self) -> np.ndarray:
        return np.c_[
            self.base_cloud.points(),
            self.base_cloud.getScalarField(
                self.base_cloud.getScalarFieldIndexByName(self.config.basic.z0_name)
            ).asArray(),
        ].astype(np.double)

    def _get_xyz_from_base(self) -> np.ndarray:
        # TODO(RJ) double conversion is only needed for DTM processing,
        # But maybe it's worth generalizing it.
        return self.base_cloud.points().astype(np.double)

    def _export_dtm(self, dtm: np.ndarray):
        cloud_dtm = pycc.ccPointCloud(dtm[:, 0], dtm[:, 1], dtm[:, 2])
        cloud_dtm.copyGlobalShiftAndScale(self.base_cloud)
        cloud_dtm.setName("dtm")
        cloud_dtm.setEnabled(False)
        self.base_group.addChild(cloud_dtm)
        self.cc_instance.addToDB(cloud_dtm)

    def _export_stripe(self, clust_stripe: np.ndarray):
        cloud_stripe = pycc.ccPointCloud(clust_stripe[:, 0], clust_stripe[:, 1], clust_stripe[:, 2])
        cloud_stripe.copyGlobalShiftAndScale(self.base_cloud)
        cloud_stripe.setName("Stems in stripe")
        CloudComparePluginProcessing.write_sf(cloud_stripe, clust_stripe[:, -1], "tree_ID")
        cloud_stripe.setCurrentDisplayedScalarField(0)
        cloud_stripe.toggleSF()
        cloud_stripe.setEnabled(False)
        self.base_group.addChild(cloud_stripe)
        self.cc_instance.addToDB(cloud_stripe)

    def _enrich_base_cloud(self, assigned_cloud: np.ndarray):
        copy_base_cloud = pycc.ccPointCloud(self.base_cloud.getName())
        copy_base_cloud.copyGlobalShiftAndScale(self.base_cloud)
        copy_base_cloud.reserve(self.base_cloud.size())

        # Could be a pycc.ccPointCloud.clone() but we do not want to clone all SFs
        copy_base_cloud.addPoints(assigned_cloud[:, 0], assigned_cloud[:, 1], assigned_cloud[:, 2])

        CloudComparePluginProcessing.write_sf(copy_base_cloud, assigned_cloud[:, 5], "dist_axes")
        CloudComparePluginProcessing.write_sf(copy_base_cloud, assigned_cloud[:, 4], "tree_ID")

        # Use assigned_cloud z0 anyway
        CloudComparePluginProcessing.write_sf(copy_base_cloud, assigned_cloud[:, 3], "Z0")

        copy_base_cloud.toggleSF()
        copy_base_cloud.setCurrentDisplayedScalarField(0)  # dist_axes

        # Set our custom color scale
        copy_base_cloud.setEnabled(False)
        color_scale_uuid = "{25ec76a1-9b8d-4e4a-a129-21ae313ef8ba}"
        color_scale_manager = pycc.ccColorScalesManager.GetUniqueInstance()
        color_scale = color_scale_manager.getScale(color_scale_uuid)
        copy_base_cloud.getCurrentDisplayedScalarField().setColorScale(color_scale)

        self.base_group.addChild(copy_base_cloud)
        self.cc_instance.addToDB(copy_base_cloud)

    def _export_tree_height(self, tree_heights: np.ndarray):
        cloud_tree_heights = pycc.ccPointCloud(tree_heights[:, 0], tree_heights[:, 1], tree_heights[:, 2])
        cloud_tree_heights.copyGlobalShiftAndScale(self.base_cloud)
        cloud_tree_heights.setName("Highest points")
        CloudComparePluginProcessing.write_sf(cloud_tree_heights, tree_heights[:, 3], "z0")
        CloudComparePluginProcessing.write_sf(cloud_tree_heights, tree_heights[:, 4], "deviated")
        cloud_tree_heights.setPointSize(8)
        z0 = cloud_tree_heights.getScalarField(0)  # z0

        # Add label with z0 values
        for i in range(len(cloud_tree_heights.points())):
            hlabel = pycc.cc2DLabel(f"point{i}")
            hlabel.addPickedPoint(cloud_tree_heights, i)
            value = round(z0.asArray()[i], 2)
            hlabel.setName(f"{value:.2f}")
            hlabel.displayPointLegend(True)
            hlabel.toggleVisibility()
            hlabel.setDisplayedIn2D(False)
            cloud_tree_heights.addChild(hlabel)
            self.cc_instance.addToDB(hlabel)

        # Set black color everywhere
        cloud_tree_heights.setColor(0, 0, 0, 255)
        cloud_tree_heights.toggleColors()
        self.base_group.addChild(cloud_tree_heights)
        self.cc_instance.addToDB(cloud_tree_heights, autoExpandDBTree=False)

    def _export_circles(self, circles_coords: np.ndarray):
        cloud_circles = pycc.ccPointCloud(circles_coords[:, 0], circles_coords[:, 1], circles_coords[:, 2])
        cloud_circles.copyGlobalShiftAndScale(self.base_cloud)
        cloud_circles.setName("Fitted sections")
        CloudComparePluginProcessing.write_sf(cloud_circles, circles_coords[:, 4], "tree_ID")
        CloudComparePluginProcessing.write_sf(cloud_circles, circles_coords[:, 5], "sector_occupancy_percent")
        CloudComparePluginProcessing.write_sf(cloud_circles, circles_coords[:, 6], "pts_inner_circle")
        CloudComparePluginProcessing.write_sf(cloud_circles, circles_coords[:, 7], "Z0")
        CloudComparePluginProcessing.write_sf(cloud_circles, circles_coords[:, 8], "Diameter")
        CloudComparePluginProcessing.write_sf(cloud_circles, circles_coords[:, 9], "outlier_prob")
        CloudComparePluginProcessing.write_sf(cloud_circles, circles_coords[:, 10], "quality")
        cloud_circles.toggleSF()
        cloud_circles.setCurrentDisplayedScalarField(6)  # = quality
        self.base_group.addChild(cloud_circles)
        self.cc_instance.addToDB(cloud_circles)

    def _export_axes(self, axes_points: np.ndarray, tilt: np.ndarray):
        cloud_axes = pycc.ccPointCloud(axes_points[:, 0], axes_points[:, 1], axes_points[:, 2])
        cloud_axes.copyGlobalShiftAndScale(self.base_cloud)
        cloud_axes.setName("Axes")
        CloudComparePluginProcessing.write_sf(cloud_axes, tilt, "tilting_degree")
        cloud_axes.toggleSF()
        cloud_axes.setCurrentDisplayedScalarField(0)  # = tilting_degree
        cloud_axes.setEnabled(False)
        self.base_group.addChild(cloud_axes)
        self.cc_instance.addToDB(cloud_axes)

    def _export_tree_locations(self, tree_locations: np.ndarray, dbh_values: np.ndarray):
        cloud_tree_locations = pycc.ccPointCloud(
            tree_locations[:, 0],
            tree_locations[:, 1],
            tree_locations[:, 2],
        )
        cloud_tree_locations.copyGlobalShiftAndScale(self.base_cloud)
        cloud_tree_locations.setName("Tree locator")
        cloud_tree_locations.setPointSize(8)
        CloudComparePluginProcessing.write_sf(cloud_tree_locations, dbh_values.reshape(-1), "dbh")
        cloud_tree_locations.setColor(255, 0, 255, 255)
        cloud_tree_locations.toggleColors()
        dbh = cloud_tree_locations.getScalarField(0)  # dbh

        for i in range(len(cloud_tree_locations.points())):
            dlabel = pycc.cc2DLabel(f"point{i}")
            dlabel.addPickedPoint(cloud_tree_locations, i)
            value = round(dbh.asArray()[i], 3)
            if value == 0.0:
                dlabel.setName(f"Tree {i+1} | Non Reliable")
            else:
                dlabel.setName(f"Tree {i+1} | {value:.3f}")
            dlabel.displayPointLegend(True)
            dlabel.toggleVisibility()
            dlabel.setDisplayedIn2D(False)
            cloud_tree_locations.addChild(dlabel)
            self.cc_instance.addToDB(dlabel)

        self.base_group.addChild(cloud_tree_locations)
        self.cc_instance.addToDB(cloud_tree_locations, autoExpandDBTree=False)

    def _export_tabular_data(
        self,
        config,
        output_basepath,
        X_c,
        Y_c,
        R,
        check_circle,
        sector_perct,
        n_points_in,
        sections,
        outliers,
        dbh_values,
        tree_locations,
        tree_heights,
        cloud_size,
        cloud_shape,
    ):
        """Revert the Global shift in tabular data when needed.

        If the base cloud was shifted, the coordinates are converted back in global ones
        before exporting the tabular data by simply calling the parent method.
        see parent method for more details on parmeters.
        """
        if self.base_cloud.isShifted():
            print("Cloud is shifted, correcting tabular data coordinates.")
            global_shift = self.base_cloud.getGlobalShift()
            global_scale = self.base_cloud.getGlobalScale()

            # Convert coordinates to global ones.
            X_c = X_c / global_scale - global_shift[0]
            Y_c = Y_c / global_scale - global_shift[1]
            R = R / global_scale
            tree_locations[:, 0] = tree_locations[:, 0] / global_scale - global_shift[0]
            tree_locations[:, 1] = tree_locations[:, 1] / global_scale - global_shift[1]
            tree_locations[:, 2] = tree_locations[:, 2] / global_scale - global_shift[2]
            tree_heights[:, 0] = tree_heights[:, 0] / global_scale - global_shift[0]
            tree_heights[:, 1] = tree_heights[:, 1] / global_scale - global_shift[1]
            tree_heights[:, 2] = tree_heights[:, 2] / global_scale - global_shift[2]

        print("Exporting tabular data")
        super()._export_tabular_data(
            config,
            output_basepath,
            X_c,
            Y_c,
            R,
            check_circle,
            sector_perct,
            n_points_in,
            sections,
            outliers,
            dbh_values,
            tree_locations,
            tree_heights,
            cloud_size,
            cloud_shape,
        )
